@@ -29,40 +29,47 @@
 */
 
 
+#pragma once
 #ifndef OPENIMAGEIO_IMAGEBUFALGO_UTIL_H
 #define OPENIMAGEIO_IMAGEBUFALGO_UTIL_H
 
-#include <OpenImageIO/platform.h>
+#include <functional>
+
+#include <OpenImageIO/parallel.h>
 #include <OpenImageIO/imagebufalgo.h>
-#include <OpenImageIO/thread.h>
 
-#if OIIO_CPLUSPLUS_VERSION >= 11
-# include <functional>
-#else
-# include <boost/bind.hpp>
-#endif
 
 
 OIIO_NAMESPACE_BEGIN
 
-#if OIIO_CPLUSPLUS_VERSION >= 11
 using std::bind;
 using std::ref;
 using std::cref;
 using namespace std::placeholders;
 using std::placeholders::_1;
-#else
-using boost::bind;
-using boost::ref;
-using boost::cref;
-using boost::this_thread::get_id;
-#endif
-
 
 
 namespace ImageBufAlgo {
 
-enum SplitDir { Split_X, Split_Y, Split_Z, Split_Biggest };
+
+/// Split strategies used by parallel_image();
+enum SplitDir { Split_X, Split_Y, Split_Z, Split_Biggest, Split_Tile };
+
+
+/// Encapsulation of options that control parallel_image().
+struct parallel_image_options {
+    parallel_image_options () {}
+    parallel_image_options (int maxthreads) : maxthreads(maxthreads) { }
+    parallel_image_options (int maxthreads, SplitDir splitdir)
+        : maxthreads(maxthreads), splitdir(splitdir) { }
+
+    int maxthreads = 0;           // Max threads (0 = use all)
+    SplitDir splitdir = Split_Y;  // Primary split direction
+    size_t minpixels = 16384;     // Min pixels per task
+    thread_pool *pool = nullptr;  // If non-NULL, custom thread pool
+};
+
+
 
 /// Helper template for generalized multithreading for image processing
 /// functions.  Some function/functor f is applied to every pixel the
@@ -79,67 +86,74 @@ enum SplitDir { Split_X, Split_Y, Split_Z, Split_Biggest };
 /// Most image operations will require additional arguments, including
 /// additional input and output images or other parameters.  The
 /// parallel_image template can still be used by employing the
-/// std::bind (or boost::bind, for C++ < 11).  For example, suppose you
-/// have an image operation defined as:
+/// std::bind. For example, suppose you have an image operation defined as:
 ///     void my_image_op (ImageBuf &out, const ImageBuf &in,
 ///                       float scale, ROI roi);
 /// Then you can parallelize it as follows:
 ///     ImageBuf R, A;   // result, input
 ///     ROI roi = get_roi (R.spec());
 ///     parallel_image (bind(my_image_op,ref(R), cref(A),3.14,_1), roi);
-///
-template <class Func>
-void
-parallel_image (Func f, ROI roi, int nthreads=0, SplitDir splitdir=Split_Y)
+inline void
+parallel_image (ROI roi, parallel_image_options opt,
+                std::function<void(ROI)> f)
 {
-    // Special case: threads <= 0 means to use the "threads" attribute
-    if (nthreads <= 0)
-        OIIO::getattribute ("threads", nthreads);
+    thread_pool *pool = opt.pool ? opt.pool : default_thread_pool();
+    // Special case: threads <= 0 means to use the pool size
+    int nthreads = (opt.maxthreads > 0) ? opt.maxthreads : pool->size();
     // Try not to assign a thread less than 16k pixels, or it's not worth
     // the thread startup/teardown cost.
-    nthreads = std::min (nthreads, 1 + int(roi.npixels() / size_t(16384)));
-    if (nthreads <= 1) {
-        // Just one thread, or a small image region: use this thread only
+    nthreads = std::min (nthreads, 1 + int(roi.npixels() / opt.minpixels));
+    if (nthreads <= 1 || pool->this_thread_is_in_pool()) {
+        // Just one thread, or a small image region, or recursive use of
+        // parallel_image: use this thread only
         f (roi);
         return;
     }
 
     // If splitdir was not explicit, find the longest edge.
+    SplitDir splitdir = opt.splitdir;
     if (splitdir >= Split_Biggest)
         splitdir = roi.width() > roi.height() ? Split_X : Split_Y;
-    int minmax[6] = { roi.xbegin, roi.xend, roi.ybegin, roi.yend,
-                      roi.zbegin, roi.zend };
-    int roi_begin = minmax[2*int(splitdir)];
-    int roi_end = minmax[2*int(splitdir)+1];
-    int splitlen = roi_end - roi_begin;
-    nthreads = std::min (nthreads, splitlen);
 
-    // Spawn threads by dividing the region into bands.
-    OIIO::thread_group threads;
-    int blocksize = std::max (1, (splitlen + nthreads - 1) / nthreads);
-    for (int i = 0;  i < nthreads;  i++) {
-        if (splitdir == Split_Y) {
-            roi.ybegin = roi_begin + i * blocksize;
-            roi.yend = std::min (roi.ybegin + blocksize, roi_end);
-            if (roi.ybegin >= roi.yend)
-                break;   // no more work to dole out
-        } else if (splitdir == Split_X) {
-            roi.xbegin = roi_begin + i * blocksize;
-            roi.xend = std::min (roi.xbegin + blocksize, roi_end);
-            if (roi.xbegin >= roi.xend)
-                break;   // no more work to dole out
-        } else { // if (splitdir == Split_Z)
-            roi.zbegin = roi_begin + i * blocksize;
-            roi.zend = std::min (roi.zbegin + blocksize, roi_end);
-            if (roi.zbegin >= roi.zend)
-                break;   // no more work to dole out
-        }
-        if (i < nthreads-1)
-            threads.add_thread (new OIIO::thread (f, roi));
-        else
-            f (roi);   // Run the last one in the calling thread
+    int64_t xchunk = 0, ychunk = 0;
+    if (splitdir == Split_Y) {
+        xchunk = roi.width();
+        // ychunk = std::max (64, minpixels/xchunk);
+    } else if (splitdir == Split_X) {
+        ychunk = roi.height();
+        // ychunk = std::max (64, minpixels/xchunk);
+    } else if (splitdir == Split_Tile) {
+        int64_t n = std::max<imagesize_t>(opt.minpixels, roi.npixels());
+        xchunk = ychunk = std::max (1, int(sqrt(n))/4);
+    } else {
+        xchunk = ychunk = std::max (int64_t(1), int64_t(sqrt(nthreads))/2);
     }
-    threads.join_all ();
+
+    auto task = [&](int id, int64_t xbegin, int64_t xend,
+                    int64_t ybegin, int64_t yend) {
+        f (ROI (xbegin, xend, ybegin, yend, roi.zbegin, roi.zend,
+                roi.chbegin, roi.chend));
+    };
+    parallel_for_chunked_2D (roi.xbegin, roi.xend, xchunk,
+                             roi.ybegin, roi.yend, ychunk, task);
+}
+
+
+inline void
+parallel_image (ROI roi, std::function<void(ROI)> f)
+{
+    parallel_image (roi, parallel_image_options(), f);
+}
+
+
+
+// DEPRECATED -- eventually enable the OIIO_DEPRECATION
+template <class Func>
+// OIIO_DEPRECATED("switch to new parallel_image")
+void
+parallel_image (Func f, ROI roi, int nthreads=0, SplitDir splitdir=Split_Y)
+{
+    parallel_image (roi, parallel_image_options (nthreads, splitdir), f);
 }
 
 
@@ -183,6 +197,7 @@ enum IBAprep_flags {
     IBAprep_DEEP_MIXED = 1<<12,         // Allow deep & non-deep combinations
     IBAprep_DST_FLOAT_PIXELS = 1<<13,   // If dst is uninit, make it float
     IBAprep_MINIMIZE_NCHANNELS = 1<<14, // Multi-inputs get min(nchannels)
+    IBAprep_REQUIRE_MATCHING_CHANNELS = 1<<15, // Channel names must match
 };
 
 
